'''
usage:
  bf cache [options] clear
  bf cache [options] compact

global options:
  -h --help                 Show help
  --dataset=<dataset>       Use specified dataset (instead of your current working dataset)
  --profile=<name>          Use specified profile (instead of default)
'''
from __future__ import absolute_import, division, print_function

from docopt import docopt

from blackfynn.cache import get_cache


def main(bf):
    args = docopt(__doc__)

    cache = get_cache(bf.settings, init=False)

    if args['clear']:
        print("Clearing cache...")
        cache.clear()
        print("Cache cleared.")
    elif args['compact']:
        print('Compacting cache...')
        cache.start_compaction(background=False)
        print('Cache compaction done.')



from __future__ import absolute_import, division, print_function

import os
from functools import reduce

from .working_dataset import working_dataset_id


def get_item_path(bf, item):
    if item.parent is not None:
        parent = bf.get(item.parent)
        return get_item_path(bf, parent) + [item]
    else:
        dataset = bf.get_dataset(item.dataset)
        return [dataset, item]

def make_tree(path):
    if len(path)==1:
        value = path[0]
    else:
        value = make_tree(path[1:])
    return {path[0].id: value}

def merge(a, b, path=None):
    if path is None: path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass # same leaf value
            else:
                raise Exception('Conflict at %s' % '.'.join(path + [str(key)]))
        else:
            a[key] = b[key]
    return a

def print_tree(tree, objects, indent=0):
    for key, value in tree.items():
        is_dict = isinstance(value, dict)
        char = '-' if is_dict else '*'
        print(' '*indent+char, objects[key])
        if is_dict:
            print_tree(value, objects, indent=indent+2)

def print_path_tree(bf, results):
    paths = [get_item_path(bf, r) for r in results]
    path_objects = {o.id:o for p in paths for o in p}
    trees = [make_tree(path) for path in paths]
    print_tree(reduce(merge, trees), path_objects)

# destination must be a full Dataset or Collection object
def recursively_upload(bf, destination, files):
    import os
    from blackfynn import Collection

    dirs = [f for f in files if os.path.isdir(f)]
    files = [f for f in files if os.path.isfile(f)]

    if len(files) > 0:
        bf._api.io.upload_files(destination, files, display_progress=True)

    for d in dirs:
        name = os.path.basename(os.path.normpath(d))
        print('Uploading to {}'.format(name))

        new_collection = Collection(name)
        destination.add(new_collection)

        files = [os.path.join(d,f) for f in os.listdir(d) if not f.startswith('.')]
        recursively_upload(bf, new_collection, files)

def print_datasets(bf):
    wd = working_dataset_id()
    for dataset in bf.datasets():
        if dataset.id == wd:
            print("\033[32m* {} (id: {})\033[0m".format(dataset.name, dataset.id))
        else:
            print("  {} (id: {})".format(dataset.name, dataset.id))

def get_item(identifier, bf):
    try:
        return bf.get(identifier)
    except:
        exit("{} does not exist.".format(identifier))



'''
usage:
  bf [options] init <name>

global options:
  -h --help                 Show help
  --profile=<name>          Use specified profile (instead of default)
'''
from __future__ import absolute_import, division, print_function

import sys

from docopt import docopt

from .cli_utils import print_datasets
from .working_dataset import set_working_dataset


def main(bf):
    args = docopt(__doc__)

    name = args['<name>']
    dataset = bf.create_dataset(name)

    set_working_dataset(dataset)

    print_datasets(bf)



'''
usage:
  bf move [options] <item> [<destination>]

global options:
  -h --help                 Show help
  --dataset=<dataset>       Use specified dataset (instead of your current working dataset)
  --profile=<name>          Use specified profile (instead of default)
'''
from __future__ import absolute_import, division, print_function

from docopt import docopt

import blackfynn.log as log
from blackfynn import Dataset
from blackfynn.models import Collection

from .cli_utils import get_item

logger = log.get_logger('blackfynn.cli.bf_move')

def main(bf):
    args = docopt(__doc__)

    item = get_item(args['<item>'], bf)

    if args['<destination>']:
        destination = get_item(args['<destination>'], bf)
    else:
        # item will be moved to to the top of its containing dataset
        destination = None

    if destination is None or isinstance(destination, Collection):
        try:
            bf.move(destination, item)
        except Exception as e:
            logger.error(e)
            exit("Failed to move {} into {}.".format(item, destination))
    else:
        exit("Destination must be a collection.")



'''
usage:
  bf where [options] <item>

global options:
  -h --help                 Show help
  --dataset=<dataset>       Use specified dataset (instead of your current working dataset)
  --profile=<name>          Use specified profile (instead of default)
'''
from __future__ import absolute_import, division, print_function



from docopt import docopt

from .cli_utils import get_item, print_path_tree


def main(bf):
    args = docopt(__doc__)

    item = get_item(args['<item>'], bf)
    print_path_tree(bf, [item])



'''
usage:
  bf append [options] <destination> <file>...

global options:
  -h --help                 Show help
  --dataset=<dataset>       Use specified dataset (instead of your current working dataset)
  --profile=<name>          Use specified profile (instead of default)
'''
from __future__ import absolute_import, division, print_function

from docopt import docopt

from blackfynn.models import DataPackage


def main(bf):
    args = docopt(__doc__)

    files = args['<file>']
    destination  = args['<destination>']

    try:
        package = bf.get(destination)

        if not isinstance(package, DataPackage):
            raise Exception("Only data packages may be appended to.")

        bf._api.io.upload_files(package, files, append=True, display_progress=True)
    except Exception as e:
        exit(e)



'''
usage:
  bf [options] ( datasets | ds )

global options:
  -h --help                 Show help
  --dataset=<dataset>       Use specified dataset (instead of your current working dataset)
  --profile=<name>          Use specified profile (instead of default)
'''
from __future__ import absolute_import, division, print_function

import sys

from docopt import docopt

from .cli_utils import print_datasets


def main(bf):
    args = docopt(__doc__)

    print_datasets(bf)



'''
usage:
  bf upload [options] [--to=destination] <file>...

global options:
  -h --help                 Show help
  --dataset=<dataset>       Use specified dataset (instead of your current working dataset)
  --profile=<name>          Use specified profile (instead of default)
'''
from __future__ import absolute_import, division, print_function

import os

from docopt import docopt

from .cli_utils import recursively_upload
from .working_dataset import require_working_dataset


def main(bf):
    args = docopt(__doc__)

    files = args['<file>']

    if args['--to']:
        collection = bf.get(args['--to'])
        recursively_upload(bf, collection, files)

    else:
        dataset = require_working_dataset(bf)
        recursively_upload(bf, dataset, files)



'''
usage:
  bf [options] ( organizations | orgs )

global options:
  -h --help                 Show help
  --profile=<name>          Use specified profile (instead of default)
'''
from __future__ import absolute_import, division, print_function

import sys

from docopt import docopt


def main(bf):
    args = docopt(__doc__)

    for o in bf.organizations():
        print("  {} (id: {})".format(o.name, o.id))



'''
usage:
  bf status [options]

global options:
  -h --help                 Show help
  --dataset=<dataset>       Use specified dataset (instead of your current working dataset)
  --profile=<name>          Use specified profile (instead of default)
'''
from __future__ import absolute_import, division, print_function

from docopt import docopt

from .working_dataset import working_dataset_id


def main(bf):
    args = docopt(__doc__)
    print('Active profile:\n  \033[32m{}\033[0m\n'.format(bf.settings.active_profile))

    if len(bf.settings.env.keys()) > 0:
        key_len = 0
        value_len = 0
        for key,value in bf.settings.env.items():
            valstr = '{}'.format(value)
            key_len = max(key_len,len(key))
            value_len = max(value_len,len(valstr))

        print('Environment variables:')
        print('  \033[4m{:{key_len}}    {:{value_len}}    {}'.format('Key','Value','Environment Variable\033[0m',key_len=key_len,value_len=value_len))
        for value,evar in sorted(bf.settings.env.items()):
           valstr = '{}'.format(value)
           # get internal variable
           ivar=''
           for k,v in bf.settings.__dict__.items():
                if str(v) == str(evar):
                    ivar = k
           print('  {:{key_len}}    {:{value_len}}    {}'.format(ivar,evar,valstr,key_len=key_len,value_len=value_len))
        print()

    working_dataset_status = 'Not set.'
    ds = working_dataset_id()
    if ds:
        try:
            working_dataset = bf.get_dataset(ds)
            working_dataset_status = "{} (id: {})".format(working_dataset.name, working_dataset.id)
        except:
            pass

    print("Blackfynn environment:")
    print("  User               : {}".format(bf.profile.email))
    print("  Organization       : {} (id: {})".format(bf.context.name, bf.context.id))
    print("  Dataset            : {}".format(working_dataset_status))
    print("  API Location       : {}".format(bf.settings.api_host))
    print("  Streaming API      : {}".format(bf.settings.streaming_api_host))
    print("  Models API         : {}".format(bf.settings.concepts_api_host))



'''
usage:
  bf create [options] <name> [<destination>]

global options:
  -h --help                 Show help
  --dataset=<dataset>       Use specified dataset (instead of your current working dataset)
  --profile=<name>          Use specified profile (instead of default)
'''
from __future__ import absolute_import, division, print_function

from docopt import docopt

from blackfynn import Collection

from .cli_utils import get_item
from .working_dataset import require_working_dataset


def main(bf):
    args = docopt(__doc__)

    collection = Collection(args['<name>'])

    if args['<destination>']:
        parent = get_item(args['<destination>'], bf)
        resp = parent.add(collection)
    else:
        dataset = require_working_dataset(bf)
        resp = dataset.add(collection)

    print(collection)



'''
usage:
  bf use [options] <dataset>

global options:
  -h --help                 Show help
  --profile=<name>          Use specified profile (instead of default)
'''
from __future__ import absolute_import, division, print_function

from docopt import docopt

from .working_dataset import set_working_dataset


def main(bf):
    args = docopt(__doc__)

    dataset_id_or_name = args['<dataset>']

    try:
        dataset = bf.get_dataset(dataset_id_or_name)
        set_working_dataset(dataset)
    except Exception as e:
        exit(e)



'''
usage:
  bf rename [options] <item> <new_name>

global options:
  -h --help                 Show help
  --dataset=<dataset>       Use specified dataset (instead of your current working dataset)
  --profile=<name>          Use specified profile (instead of default)
'''
from __future__ import absolute_import, division, print_function

from docopt import docopt

from blackfynn.models import BaseDataNode

from .cli_utils import get_item


def main(bf):
    args = docopt(__doc__)

    item = get_item(args['<item>'], bf)
    item.name = args['<new_name>']
    item.update()



'''
usage:
  bf collaborators [options]
  bf share [options] <collaborators>...
  bf unshare [options] <collaborators>...

global options:
  -h --help                 Show help
  --dataset=<dataset>       Use specified dataset (instead of your current working dataset)
  --profile=<name>          Use specified profile (instead of default)
'''
from __future__ import absolute_import, division, print_function

import sys


from docopt import docopt

from .working_dataset import require_working_dataset


def main(bf):
    args = docopt(__doc__)

    ds = require_working_dataset(bf)

    if args['collaborators']:
        print("'bf collaborators' is deprecated since version 2.7.2 and will be removed")
        print("Manage permissions through the Blackfynn web app instead\n")

        resp = ds.collaborators

        if resp['users'] or resp['organizations']:
            print(" - Users")
            for u in resp['users']:
                print("   - email:{} id:{}".format(u.email, u.id))

            print(" - Organizations")
            for o in resp['organizations']:
                print("   - name:{} id:{}".format(o.name, o.id))
        else:
            print("  No collaborators.")

    elif args['share']:
        print("'bf share' is deprecated since version 2.7.2 and will be removed")
        print("Manage permissions through the Blackfynn web app instead\n")

        ids = args['<collaborators>']

        resp = ds.add_collaborators(*ids)
        print_collaborator_edit_resp(resp)

    elif args['unshare']:
        print("'bf unshare' is deprecated since version 2.7.2 and will be removed")
        print("Manage permissions through the Blackfynn web app instead\n")

        ids = args['<collaborators>']

        resp = ds.remove_collaborators(*ids)
        print_collaborator_edit_resp(resp)

def print_collaborator_edit_resp(resp):
    for key, value in list(resp['changes'].items()):
        if value['success']:
            print(" - {}: Success".format(key))
        else:
            print(" - {}: Error - {}".format(key, value['message']))



from __future__ import absolute_import, division, print_function

import io
import os
import tempfile

import psutil


def set_working_dataset(dataset):
    ds_file = get_dataset_file()
    try:
        with io.open(ds_file, 'w+') as f:
            f.write(dataset.id)
        # double check
        assert working_dataset_id() is not None, "Error writing to working dataset file"
    except:
        print("We encountered an error while setting your working dataset.")
    return None

def get_dataset_file():
    try:
        if os.name == 'nt':
            parent_shell_pid = psutil.Process(os.getpid()).parent().parent().pid
        else:
            parent_shell_pid = psutil.Process(os.getpid()).parent().pid
    except:
        return None
    fname = 'blackfynn_working_dataset_{}'.format(parent_shell_pid)
    return os.path.join(tempfile.gettempdir(), fname)

def working_dataset_id():
    ds_file = get_dataset_file()
    if os.path.exists(ds_file):
        try:
            with io.open(ds_file, 'r+') as f:
                ds_id = f.read().strip()
                return ds_id
        except:
            pass
    return None

def require_working_dataset(bf):
    ds = working_dataset_id()
    if ds is None:
        exit("Dataset required. You can set your working dataset with:" \
             "\n\n\tbf use <dataset>" \
             "\n\nor by specifying it in your command:" \
             "\n\n\t bf --dataset=<dataset> <command> ..." \
             "\n")
    try:
        return bf.get_dataset(ds)
    except Exception as e:
        exit(e)



'''Describe a package or collection

usage:
  bf get [options] [--tree] [<item>]

options:
  --tree                    Recursively print the item's contents

global options:
  -h --help                 Show help
  --dataset=<dataset>       Use specified dataset (instead of your current working dataset)
  --profile=<name>          Use specified profile (instead of default)
'''
from __future__ import absolute_import, division, print_function

from docopt import docopt

from .cli_utils import get_item
from .working_dataset import require_working_dataset


def display(item, print_tree):
    if print_tree:
        item.print_tree()

    else:
        print("  {}".format(item))

        if hasattr(item, 'items') and item.items:
            print("  [contents]")
            for i in item.items:
                print("     {}".format(i))

        if hasattr(item, 'channels') and item.channels:
            print("  [channels]")
            for ch in item.channels:
                print("     {} (id: {})".format(ch.name, ch.id))

def main(bf):
    args = docopt(__doc__)

    if args['<item>']:
        item = get_item(args['<item>'], bf)
    else:
        item = require_working_dataset(bf)

    display(item, args['--tree'])
